import ast


class EnvVisitor(ast.NodeVisitor):
    def __init__(self):
        self.optional_environment_variables = set()
        self.required_environment_variables = set()

    def parse_and_visit(self, body, filename=''):
        doc = ast.parse(body, filename=filename)
        return self.visit(doc)

    def visit_Call(self, call):
        is_getenv = False
        is_environ_get = False
        if isinstance(call.func, ast.Attribute):
            if call.func.attr == 'getenv':
                is_getenv = True
            elif call.func.attr == 'get':
                if isinstance(call.func.value, ast.Attribute) and call.func.value.attr == 'environ':
                    is_environ_get = True
                elif isinstance(call.func.value, ast.Name) and call.func.value.id == 'environ':
                    is_environ_get = True
        elif isinstance(call.func, ast.Name):
            if call.func.id == 'getenv':
                is_getenv = True
        if is_getenv:
            if len(call.args) >= 1 and isinstance(call.args[0], ast.Str):
                self.optional_environment_variables.add(ast.literal_eval(call.args[0]))
        elif is_environ_get:
            if len(call.args) >= 1 and isinstance(call.args[0], ast.Str):
                self.optional_environment_variables.add(ast.literal_eval(call.args[0]))
        self.generic_visit(call)

    def visit_Subscript(self, what):
        is_env_slice = False
        if isinstance(what.value, ast.Attribute) and what.value.attr == 'environ':
            is_env_slice = True
        elif isinstance(what.value, ast.Name) and what.value.id == 'environ':
            is_env_slice = True
        if is_env_slice:
            if isinstance(what.slice, ast.Index) and isinstance(what.slice.value, ast.Str):
                self.required_environment_variables.add(ast.literal_eval(what.slice.value))
        self.generic_visit(what)



from guessenv import guesser


def test_basic():
    e = guesser.EnvVisitor()
    e.parse_and_visit('''
    import os
    from os import environ

    os.environ["one"]
    os.environ.get("two")
    os.environ.get("three", default="bar")
    os.getenv("four")
    os.getenv("five", "bar")
    environ['six']
    environ['sev' 'en']

    not_environment = {}
    not_environment.get('nope')
    not_environment['nope']
    '''.replace('    ', ''))
    assert e.optional_environment_variables == set(['two', 'three', 'four', 'five'])
    assert e.required_environment_variables == set(['one', 'six', 'seven'])


def test_environ_method_call():
    e = guesser.EnvVisitor()
    e.parse_and_visit('''
    import os

    KEY = os.environ['foo'].strip()
    '''.replace('    ', ''))
    assert e.optional_environment_variables == set()
    assert e.required_environment_variables == set(['foo'])


def test_nested():
    e = guesser.EnvVisitor()
    e.parse_and_visit('''
    import os

    FOO = os.environ[os.environ['foo']]
    BAR = os.getenv(os.getenv('bar'))
    '''.replace('    ', ''))
    assert e.optional_environment_variables == set(['bar'])
    assert e.required_environment_variables == set(['foo'])



#!/usr/bin/env python

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils import setup, find_packages


setup(
    name='guessenv',
    version='0.2',
    author='EasyPost OSS',
    author_email='oss@easypost.com',
    url='https://github.com/easypost/guessenv',
    license='ISC',
    packages=find_packages(exclude=['test']),
    description='Parse Python files and look for required environment variables',
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    entry_points={
        'console_scripts': [
            'guessenv = guessenv.__main__:main',
        ]
    },
    project_urls={
        'Issue Tracker': 'https://github.com/easypost/guessenv/issues',
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Intended Audience :: End Users/Desktop',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: ISC License (ISCL)',
    ]
)



import argparse
import sys
import io
import os
import os.path
import re
from itertools import repeat

from . import guesser


if sys.version_info < (3, 0):
    open_function = io.open
else:
    open_function = open


class _Exclude(object):
    def __init__(self, regex):
        self.raw = regex
        self.path = re.compile(regex)

    def match(self, path):
        path = path.lstrip('./')
        return self.path.search(path)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-q', '--quiet', action='store_true')
    parser.add_argument('-O', '--with-optional', action='store_true', help='Include optional environment variables')
    parser.add_argument('-A', '--assert-present', action='store_true',
                        help='Exit non-zero if any required environment variables are missing')
    parser.add_argument('-x', '--exclude', action='append', default=[], type=_Exclude,
                        help='Exclude files matching the given regex. May be repeated')
    parser.add_argument('-X', '--standard-exclude', action='store_true', help='Add a standard set of excludes')
    parser.add_argument('files', nargs='+')
    args = parser.parse_args()

    if args.verbose and args.quiet:
        parser.error('cannot pass both --verbose and --quiet')

    if args.standard_exclude:
        args.exclude.append(_Exclude('^venv/.*'))
        args.exclude.append(_Exclude('^.tox/.*'))
        args.exclude.append(_Exclude('^test/.*'))
        args.exclude.append(_Exclude('^tests/.*'))

    filenames = []
    envvars = []

    # Handle argv.
    for file_or_dir_name in args.files:
        if os.path.isdir(file_or_dir_name):
            for root, _, files in os.walk(file_or_dir_name):
                for f in files:
                    if f.endswith('.py'):
                        full_path = os.path.join(root, f)
                        if any(exclude.match(full_path) for exclude in args.exclude):
                            continue
                        filenames.append(os.path.join(root, f))
        elif os.path.exists(file_or_dir_name):
            if not any(exclude.match(file_or_dir_name) for exclude in args.exclude):
                filenames.append(file_or_dir_name)
        else:
            pass

    status = 0
    for filename in filenames:
        v = guesser.EnvVisitor()
        with open_function(filename, 'r', encoding='utf-8') as f:
            try:
                v.parse_and_visit(f.read(), filename)
                if args.with_optional:
                    envvars.extend(list(zip(
                        repeat(filename),
                        repeat('optional'),
                        sorted(v.optional_environment_variables)
                    )))
                envvars.extend(list(zip(
                    repeat(filename),
                    repeat('required'),
                    sorted(v.required_environment_variables)
                )))
            except SyntaxError as e:
                print('{filename}:{lineno} - syntax error!'.format(
                    filename=filename,
                    lineno=e.lineno
                ))
                status = 1

    if args.assert_present:
        required_envvars = set(v for _, k, v in envvars if k == 'required')
        for envvar in required_envvars:
            if envvar not in os.environ:
                print('MISSING: {0}'.format(envvar))
                status = 1

    if args.verbose:
        for filename, kind, variable_name in envvars:
            print('{filename} {kind} {variable_name}'.format(filename=filename, kind=kind, variable_name=variable_name))
    elif not args.quiet:
        envvars = set(v for _, _, v in envvars)
        print('\n'.join(sorted(str(v) for v in envvars)))

    return status


if __name__ == '__main__':
    sys.exit(main())
